/* base32encoding/BinaryToBase32enc.c     version 2003-07-05
   Binary octets stream to base32 encoding.
   In standard C language.
   By Jean-Pierre Moreau, jpm-phys@iquebec.com.
*/

/* base32 encoding as in IETF draft "draft-josefsson-base-encoding-04.txt":
     Internet-Draft      February 2002  Expires: August 2, 2002
     Base Encoding of Data    draft-josefsson-base-encoding-04
   currently (2003-07) found on the Internet for example in
     http://www.ietf.org/internet-drafts/draft-josefsson-base-encoding-04.txt
*/

/* Input :
     stdin, 0 to nOctetsIn octets (or bytes = octet of 8 bits);
   Output :
     stdout, ( (8 * nOctetsIn)/5 + possible padding) characters, 5 bits/char.

   Padding required (with '=' char) following proposed standard of output
   in integral multiples of 'standard base32 quantum of data' =  40 bits, of
   5 input octets, so of 8 output characters.

   If no such 'multiple of base32 standard quantum' needed, then no padding
   required and we have this table of nb of out chars vs the nb of input
   octets (where m is any integer) :

     nOctetsIn     bits    nb of 5 bits needed
                            = nb of output chars

      5m + 0     40m +  0         8m + 0
      5m + 1     40m +  8         8m + 2
      5m + 2     40m + 16         8m + 4
      5m + 3     40m + 24         8m + 5
      5m + 4     40m + 32         8m + 7
*/

/* The algorithm 

  This algorithm mostly make 5 bits chunks from the inputs of 8 bits
  chunks, and pads to 'quantum of 40 bits' if not integral number of
  quantums from the input.

  Here nibbles have 5 bits, and Most Significant bit = leftmost bit of
  nibble or octet is numbered bit 0, least significant bit is bit 4 for
  nibble, bit 7 for octet.

  Here for ONE quantum for 0 (if EOF) to 5 input octets. We repeat quantums
  as needed for any number of input characters.

  1st octet :

    Read next input octet, ioctet
    If EOF, all done, integral nb of quantums, nothing for THIS quantum :
       NO MORE output, exit.
    If not EOF              (1st input octet of quantum)
       nibble0 = 5 left bits of ioctet
       nibble1(bits 0 to 2 ) = 3 right bits of ioctet

  2nd octet :

    Read next input octet, ioctet
    If EOF :
       nibble1(bits 3,4) = 0 0;
       Output this quantum ( 8 chars of output per quantum) :
         char(nibble0), char(nibble1), =,=,=,=,=,=
       NO more output, exit.
    If not EOF              (2nd input octet of quantum)
       nibble1(bits 3,4) = 2 left bits of ioctet
       nibble2           = bits 2 to 6 of ioctet
       nibble3(bit 0)    = bit 7 of ioctet = last bit of ioctet

  3rd octet :

    Read next input octet, ioctet
    If EOF :
       nibble3(bits 1 to 4) = 0 0 0 0;
       Output this quantum ( 8 chars of output per quantum) :
         char(nibble0), char(nibble1), char(nibble2), char(nibble3), =,=,=,=
       NO more output, exit.
    If not EOF              (3rd input octet of quantum)
       nibble3(bits 1 to 4) = 4 left bits of ioctet
       nibble4(bits 0 to 3) = 4 right bits of ioctet

  4th octet :

    Read next input octet, ioctet
    If EOF :
       nibble4(bits 4) = 0
       Output this quantum (8 chars of output per quantum) :
         char(nibble0),char(nibble1),...,char(nibble4),'=','=','='
       NO more output, exit.
    If not EOF              (4th input octet of quantum)
       nibble4(bit 4) = 1st left bit of ioctet
       nibble5 = bits 1 to 5 of ioctet
       nibble6(bits 0,1) = bits 6,7 of ioctet

  5th octet :

    Read next input octet, ioctet
    If EOF :
       nibble6(bits 2 to 4) = 0 0 0
       Output this quantum (8 chars of output per quantum) :
         char(nibble0),char(nibble1),...,char(nibble6), '='
       NO more output, exit.
    If not EOF              (5th input octet of quantum)
       nibble6(bits 2 to 4) = 3 left bits of ioctet
       nibble7 = 5 right bits of ioctet
    Output this quantum, complete from 5 input octets so no padding:
       char(nibble0),char(nibble1),...,char(nibble7).

    Proceed to another quantum.
*/

#include <stdio.h>
void PrintEncodeQuantumAndPadTo8(
        unsigned int nNibbles, unsigned char *theNibbles );

/* ============================= main(..) ========================== */

int main( int argc, char **argv)

{
int ioctetIn; unsigned char ioctet;

/* The 8 nibbles of 5 bits in a 40 bit quantum:
   Each nibble has its 5 bits as 5 right bits of an 8 bit unsigned char,
   so 3 left bits (most significant bits) are always 0) */
unsigned char qnibble[8];

if(argc != 1)
  { printf("Usage example : thisProg < fileOfInputOctets > outputFile\n");
    exit(1);  }

/* Proceed as many quantum (of 5 input octets) as needed */

while(1)
  {

  /* Proceed one quantum, up to 5 input octets;
     or 0 to 4 octet if End Of File (EOF) encountered */

  /* 1st octet in */

  ioctetIn=getchar();
  if(ioctetIn == EOF) /* All done, integral nb of quantums, nothing more */
      {
      exit();
      }
  ioctet = (unsigned char)ioctetIn;
  /* nibble0 = 5 left bits of ioctet;
       ioctet  :  a b c d e x y z
       nibble0 :  0 0 0 a b c d e
     nibble1(bits 0 to 2 ) = 3 right bits of ioctet :
       ioctet  :  . . . . . x y z
       nibble1 :  0 0 0 x y z 0 0                      */
  qnibble[0] = (ioctet >> 3);
  qnibble[1] = (ioctet << 2) & '\x1C';  /* ... & binary 0001 1100 */

  /* 2nd octet in */

  ioctetIn=getchar();
  if(ioctetIn == EOF)
     {
     /* Output this quantum (padded) and exit */
     PrintEncodeQuantumAndPadTo8(2,qnibble);
     exit(0);
     }
  ioctet = (unsigned char)ioctetIn;
  /* nibble1(bits 3,4) = 2 left bits of ioctet
       nibble1 (before) :  0 0 0 a b c . .
       ioctet           :  x y . . . . . .
       nibble1 (new)    :  0 0 0 a b c x y
     nibble2           = bits 2 to 6 of ioctet
       ioctet  :  . . a b c d e .
       nibble2 :  0 0 0 a b c d e
     nibble3(bit 0)    = bit 7 of ioctet = last bit of ioctet
       ioctet  :  . . . . . . . x
       nibble3 :  0 0 0 x 0 0 0 0
  */
  qnibble[1] |= (ioctet >> 6);
  qnibble[2] =  (ioctet >> 1) & '\x1F';           
  qnibble[3] =  (ioctet << 4) & '\x10';

  /* 3rd octet in */

  ioctetIn=getchar();
  if(ioctetIn == EOF)
     {
     /* Output this quantum (padded) and exit */
     PrintEncodeQuantumAndPadTo8(4,qnibble);
     exit(0);
     }
  ioctet = (unsigned char)ioctetIn;
  /* nibble3(bits 1 to 4) = 4 left bits of ioctet
       nibble3 (before) :  0 0 0 x 0 0 0 0
       ioctet           :  a b c d . . . .
       nibble3 (new)    :  0 0 0 x a b c d
     nibble4(bits 0 to 3) = 4 right bits of ioctet
       ioctet  :  . . . . p q r s
       nibble4 :  0 0 0 p q r s 0
  */
  qnibble[3] |= (ioctet >> 4);
  qnibble[4] =  (ioctet << 1) & '\x1E';           

  /* 4th octet in */

  ioctetIn=getchar();
  if(ioctetIn == EOF)
     {
     /* Output this quantum (padded) and exit */
     PrintEncodeQuantumAndPadTo8(5,qnibble);
     exit(0);
     }
  ioctet = (unsigned char)ioctetIn;
  /* nibble4(bit 4) = 1st left bit of ioctet
       nibble4 (before) :  0 0 0 p q r s 0
       ioctet           :  x . . . . . . .
       nibble4 (new)    :  0 0 0 p q r s x
     nibble5 = bits 1 to 5 of ioctet
       ioctet  :  . a b c d e . .
       nibble5 :  0 0 0 a b c d e
     nibble6(bits 0,1) = bits 6,7 of ioctet
       ioctet  :  . . . . . . t u
       nibble6 :  0 0 0 t u 0 0 0
  */
  qnibble[4] |= (ioctet >> 7);
  qnibble[5] =  (ioctet >> 2) & '\x1F';           
  qnibble[6] =  (ioctet << 3) & '\x18';

  /* 5th octet in */

  ioctetIn=getchar();
  if(ioctetIn == EOF)
     {
     /* Output this quantum (padded) and exit */
     PrintEncodeQuantumAndPadTo8(7,qnibble);
     exit(0);
     }
  ioctet = (unsigned char)ioctetIn;
  /* nibble6(bits 2 to 4) = 3 left bits of ioctet
       nibble6 (before) :  0 0 0 t u 0 0 0
       ioctet           :  a b c . . . . .
       nibble6 (new)    :  0 0 0 t u a b c
     nibble7 = 5 right bits of ioctet
       ioctet           :  . . . d e f g h
       nibble7 (new)    :  0 0 0 d e f g h
  */
  qnibble[6] |= (ioctet >> 5);
  qnibble[7] =  ioctet & '\x1F';

  /* Output this quantum, completed from 5 input octets without EOF */
  PrintEncodeQuantumAndPadTo8(8,qnibble);

  /* NOW proceed to another quantum */
  }
}
/* End of main(...) */


/* =================== PrintEncodeQuantumAndPadTo8(...) =================== */


void PrintEncodeQuantumAndPadTo8(
        unsigned int nNibbles, unsigned char *theNibbles )

{
/* The base 32 alphabet for this 'standard' encoding :

   0 to 25 is encoded to letters A to Z, than 26 to 31 to digits 2 to 7,
   so no confusion between digit 0 and number O, and between letter l and 1.

   Details: 
        (5 bit chunk) =  0  is encoded to character 'A'
        to
        (5 bit chunk) = 25  is encoded to character 'Z'
 
        (5 bit chunk) = 26  is encoded to character '2'
        to
        (5 bit chunk) = 31  is encoded to character '7'.
*/
char encod[32] = {'A','B','C','D','E','F','G','H','I',
                  'J','K','L','M','N','O','P','Q','R',
                  'S','T','U','V','W','X','Y','Z','2',
                  '3','4','5','6','7'};
/* String to print, a C string, does NOT contain '\0' except at end */
char outQuantum[9];
int i;

if( nNibbles > 8 )
    {
    fprintf(stderr,"PrintEncodeQuantumAndPadTo8: nNibbles > 8. Exit.\n");
    exit(1);
    }

for(i=0; i<nNibbles; i++)
    {
    unsigned char chunkvalue;

    if( (chunkvalue=theNibbles[i]) > 31 )
      {
      fprintf(stderr,
              "PrintEncodeQuantumAndPadTo8(..): (5 bits) value > 31. Exit.\n");
      exit(1);
      }
    outQuantum[i] = encod[chunkvalue];
    }

/* Pad remaining (8 - nNibbles) chars to padding character '=' */

for(i=nNibbles; i < 8; i++ )
    {
    outQuantum[i] = '=';
    }

outQuantum[8] = '\0'; /* for printing function, end of string */

fputs(outQuantum,stdout);

return;
}
/* End of PrintEncodeQuantumAndPadTo8(...) */
